

import os
import subprocess

from .shared import CmdError


DRY_RUN = os.environ.get("RECOMI_DRY_RUN", "FALSE").lower().startswith("t")


class Repo:

    def _run(self, cwd, args):
        if DRY_RUN:  # pragma: no cover
            print("I would have run %r" % args)
            return
        if not os.access(cwd, os.R_OK|os.W_OK|os.X_OK):
            # Subprocess will raise PermissionError if it cannot chdir to cwd
            raise CmdError("Insufficient permissions on %s" % cwd)
        env = dict(os.environ)
        env["GIT_TERMINAL_PROMPT"] = "0"
        return_code = subprocess.call(args, cwd=cwd, env=env, stderr=subprocess.STDOUT)
        if return_code != 0:
            raise CmdError("Failed")


class LocalGitRepo(Repo):

    def __init__(self, path):
        self.path = path
        self.name = self.work_out_name()

    def work_out_name(self):
        name = os.path.basename(self.path)
        if name.endswith(".git"):
            name = name[:-4]
        return name

    def is_a_bare_repo(self):
        if os.path.exists(os.path.join(self.path, ".git")):
            return False
        else:
            return True

    def fetch(self):
        args = [
            "git",
            "-c", "gc.autoDetach=false",
            "-c", "fetch.fsckObjects=true",
        ]
        if self.is_a_bare_repo():
            args.extend(["fetch", "--all"])
        else:
            args.extend(["pull"])
        self._run(self.path, args)

    def gc(self):
        args = ["git",
                "-c", "gc.autoDetach=false",
                "-c", "gc.auto=1",  # git defaults to 6700
                "gc", "--auto"]
        self._run(self.path, args)

    def fsck(self):
        args = ["git",
                "fsck", "--full"]
        self._run(self.path, args)


class UpstreamGitRepo(Repo):

    def __init__(self, name, clone_from, repo_type):
        self.name = name
        self.clone_from = clone_from
        self.repo_type = repo_type

    def dest(self):
        if self.repo_type == "working":
            return self.name
        return self.name + ".git"

    def clone_args(self):
        args = ["git", "clone"]
        if self.repo_type in ("mirror", "mirror-ff"):
            args.append("--mirror")
        if self.repo_type == "bare":
            args.append("--bare")
        args.append(self.clone_from)
        args.append(self.dest())
        return args

    def clone(self, parent):
        self._run(parent, self.clone_args())
        if self.repo_type == "mirror-ff":
            args = ["git", "config", "remote.origin.fetch", "refs/*:refs/*"]
            self._run(os.path.join(parent, self.dest()), args)





import unittest
from unittest import mock

from recomi import main


class TestOptionParsing(unittest.TestCase):

    def test_no_args(self):
        with mock.patch("argparse.ArgumentParser.error") as error_handler:
            main.parse_args([])
        error_handler.assert_called_once()

    # Hmm... fine on Py 3.11, but strange error with 3.8
    #def test_invalid_command(self):
    #    with mock.patch("argparse.ArgumentParser.error") as error_handler:
    #        main.parse_args(["notacommand", "."])
    #    error_handler.assert_called()

    def test_fetch(self):
        opts = main.parse_args(["fetch", "."])
        self.assertTrue(opts.command.should_fetch)
        self.assertFalse(opts.command.should_gc)
        self.assertFalse(opts.command.should_fsck)
        self.assertFalse(opts.command.should_clone)

    def test_gc(self):
        opts = main.parse_args(["gc", "."])
        self.assertFalse(opts.command.should_fetch)
        self.assertTrue(opts.command.should_gc)
        self.assertFalse(opts.command.should_fsck)
        self.assertFalse(opts.command.should_clone)

    def test_fsck(self):
        opts = main.parse_args(["fsck", "."])
        self.assertFalse(opts.command.should_fetch)
        self.assertFalse(opts.command.should_gc)
        self.assertTrue(opts.command.should_fsck)
        self.assertFalse(opts.command.should_clone)

    def test_clone(self):
        opts = main.parse_args(["clone", "."])
        self.assertFalse(opts.command.should_fetch)
        self.assertFalse(opts.command.should_gc)
        self.assertFalse(opts.command.should_fsck)
        self.assertTrue(opts.command.should_clone)

    def test_mirror(self):
        opts = main.parse_args(["mirror", "."])
        self.assertTrue(opts.command.should_fetch)
        self.assertTrue(opts.command.should_gc)
        self.assertFalse(opts.command.should_fsck)
        self.assertTrue(opts.command.should_clone)





import os
import unittest

from recomi import collecting
from recomi.shared import CmdError


class TestPathToName(unittest.TestCase):

    def test_ordinary(self):
        self.assertEqual(collecting.path_to_name("foo"), "foo")

    def test_bare(self):
        self.assertEqual(collecting.path_to_name("foo.git"), "foo")

    def test_path(self):
        self.assertEqual(collecting.path_to_name("baz/bar/foo.git"), "foo")
        self.assertEqual(collecting.path_to_name("baz/bar/foo.git", onlybase=False),
                         "baz__bar__foo")

    def test_ssh_host_without_path(self):
        self.assertEqual(collecting.path_to_name("bar:foo.git"), "foo")

    def test_ssh_host_and_path(self):
        self.assertEqual(collecting.path_to_name("baz:bar/foo.git"), "foo")
        self.assertEqual(collecting.path_to_name("baz:bar/foo.git", onlybase=False),
                         "bar__foo")

    def test_url(self):
        self.assertEqual(collecting.path_to_name("git://bar/foo.git"), "foo")
        self.assertEqual(collecting.path_to_name("git://bar/foo.git", onlybase=False),
                         "bar__foo")

    def test_slash_prefix(self):
        self.assertEqual(collecting.path_to_name("/bar/foo.git", onlybase=False),
                         "bar__foo")


class TestMissingCollection(unittest.TestCase):

    def test_cannot_be_instantiated(self):
        with self.assertRaises(ValueError):
            collecting.Collection("/path/to/collection/that/doesnt/exist")


class TestDummyCollectionWithConfig(unittest.TestCase):

    def setUp(self):
        dummy = os.path.join(os.path.dirname(__file__), "DummyCollection")
        self.collection = collecting.Collection(dummy)

    def test_clone_checking(self):
        self.assertEqual(self.collection.clone_not_configured(), [])

    def test_upstream_list(self):
        names = self.collection.upstream_list()
        self.assertEqual(names, ["one", "two", "three"])

    def test_url_for(self):
        expected = "https://example.com/foo/bar/reponame.git"
        self.assertEqual(self.collection.url_for("reponame", "zzz/reponame"), expected)

    def test_type(self):
        self.assertEqual(self.collection.repo_type(), "mirror")

    def test_warn_of_new_clone(self):
        self.assertEqual(self.collection.warn_of_new_clone(), True)

    def test_upstream_repos(self):
        repos = list(self.collection.upstream_repos())
        self.assertEqual(len(repos), 3)
        self.assertEqual(repos[0].name, "one")
        self.assertEqual(repos[2].name, "three")

    def test_broken_upstream_repos(self):
        self.collection.config.set("clone", "list", "exit 19")
        with self.assertRaises(CmdError) as errctx:
            list(self.collection.upstream_repos())
        self.assertIn("Exit code 19", repr(errctx.exception))





import argparse
import sys
import time

from .shared import CmdError

#from . import git
from . import collecting
from . import __version__


DESCRIPTION = "Repository Collection Mirror (version %s)" % __version__


def info(msg):  # pragma: no cover
    sys.stdout.write("%s\n" % msg)
    sys.stdout.flush()


def warn(msg):  # pragma: no cover
    if sys.stderr.isatty():
        sys.stderr.write("\033[31;1mWARNING:\033[0m recomi: %s\n" % msg)
    else:
        sys.stderr.write("WARNING: recomi: %s\n" % msg)
    sys.stderr.flush()


class LoopingCommand:

    def __init__(self, fetch=False, gc=False, fsck=False, clone=False):
        self.should_fetch = fetch
        self.should_gc = gc
        self.should_fsck = fsck
        self.should_clone = clone

    def run(self, opts):
        start_run = time.time()
        failures = 0
        for collection in opts.collections:
            info("\nCollection: %s" % collection.base_path)
            for repo in collection.local_repos():
                try:
                    if self.should_fetch:
                        info("\n[BEGIN fetch %s]" % repo.path)
                        repo.fetch()
                        info("[END fetch]")
                    if self.should_gc:
                        info("\n[BEGIN gc %s]" % repo.path)
                        repo.gc()
                        info("[END gc]")
                    if self.should_fsck:
                        info("\n[BEGIN fsck %s]" % repo.path)
                        repo.fsck()
                        info("[END fsck]")
                except CmdError as exc:
                    warn("%s: %r" % (repo.path, exc))
                    failures = failures + 1
            if self.should_clone:
                errors = collection.clone_not_configured()
                if len(errors) > 0:
                    warn("Invalid configuration for clone: %r" % errors)
                    failures = failures + 1
                    continue
                try:
                    missing_repos = list(collection.missing_repos())
                except CmdError as exc:
                    warn("list: %r" % exc)
                    failures = failures + 1
                    break
                for repo in missing_repos:
                    try:
                        info("\n[BEGIN clone %s]" % repo.clone_from)
                        if collection.warn_of_new_clone():
                            warn("New repository detected: %s" % repo.clone_from)
                        repo.clone(collection.base_path)
                        info("[END clone]")
                    except CmdError as exc:
                        warn("%s: %r" % (repo.clone_from, exc))
                        failures = failures + 1
        duration = time.time() - start_run
        info("\nRecomi finished in %i seconds" % duration)
        if failures:
            warn("================================")
            if failures == 1:
                warn("Overall there was %i failure" % failures)
            else:
                warn("Overall there were %i failures" % failures)
            sys.exit(1)


def command(value):
    """Available commands: fetch, gc, fsck, clone, mirror"""
    # Keep the docstring above up-to-date: it's used in help messages
    if value == "fetch":
        return LoopingCommand(fetch=True)
    elif value == "gc":
        return LoopingCommand(gc=True)
    elif value == "fsck":
        return LoopingCommand(fsck=True)
    elif value == "clone":
        return LoopingCommand(clone=True)
    elif value == "mirror":
        return LoopingCommand(fetch=True, gc=True, clone=True)
    else:
        raise ValueError("No such command %r" % value)


def parse_args(args):
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument("command", type=command, help=command.__doc__)
    parser.add_argument("collections", nargs="+", type=collecting.Collection)
    opts = parser.parse_args(args)
    return opts


def main():  # pragma: no cover
    opts = parse_args(sys.argv[1:])
    opts.command.run(opts)


if __name__ == "__main__":  # pragma: no cover
    main()





import unittest

from recomi import git, shared


class TestBase(unittest.TestCase):

    def test_running_good(self):
        repo = git.Repo()
        repo._run(".", ["/bin/true"])

    def test_running_bad(self):
        repo = git.Repo()
        with self.assertRaises(shared.CmdError):
            repo._run(".", ["/bin/false"])

    def test_running_invalid_directory(self):
        repo = git.Repo()
        with self.assertRaises(shared.CmdError):
            repo._run("/i/do/not/exist", ["/bin/true"])


class TestLocal(unittest.TestCase):

    def test_bare_name(self):
        repo = git.LocalGitRepo("/path/to/repo.git")
        self.assertEqual(repo.name, "repo")

    def test_name(self):
        repo = git.LocalGitRepo("/path/to/repo")
        self.assertEqual(repo.name, "repo")


class TestUpstream(unittest.TestCase):

    def test_dest_for_default_repo(self):
        repo = git.UpstreamGitRepo("foobar", "host:user/foobar.git", "mirror")
        self.assertEqual(repo.dest(), "foobar.git")

    def test_dest_for_working_repo(self):
        repo = git.UpstreamGitRepo("foobar", "host:user/foobar.git", "working")
        self.assertEqual(repo.dest(), "foobar")





import configparser
import os
import subprocess

from .shared import CmdError

from . import git


def path_to_name(path, onlybase=True):
    if ":" in path:
        ssh_host, colon, path = path.partition(":")
    if onlybase:
        name = os.path.basename(path)
    else:
        name = path.lstrip("/").replace("/", "__")
    if name.endswith(".git"):
        name = name[:-4]
    return name


class Collection:

    def __init__(self, base_path):
        if not os.path.isdir(base_path):
            raise ValueError("Invalid base directory: %r" % base_path)
        self.base_path = base_path
        self._parse_config()

    def _parse_config(self):
        self.config = configparser.ConfigParser()
        config_path = os.path.join(self.base_path, "recomi.ini")
        if os.path.exists(config_path):
            self.config.read(config_path)

    def clone_not_configured(self):
        if "clone" not in self.config.sections():  # pragma: no cover
            return ["Missing [clone] section"]
        errors = []
        if "list" not in self.config["clone"]:  # pragma: no cover
            errors.append("Missing [clone].list option")
        if "url" not in self.config["clone"]:  # pragma: no cover
            errors.append("Missing [clone].url option")
        return errors

    def local_repos(self):
        for name in sorted(os.listdir(self.base_path)):
            path = os.path.join(self.base_path, name)
            if os.path.isdir(path):
                yield git.LocalGitRepo(path)

    def upstream_list(self):
        cmd = self.config["clone"]["list"]
        try:
            output = subprocess.check_output(cmd, shell=True, cwd=self.base_path)
        except subprocess.CalledProcessError as exc:
            raise CmdError("Exit code %i from %s" % (exc.returncode, cmd))
        return output.decode("ascii").splitlines()

    def url_for(self, name, path):
        pattern = self.config["clone"]["url"]
        return pattern.format(name=name, path=path)

    def upstream_repos(self):
        onlybase = self.config["clone"].getboolean("onlybase", True)
        for path in self.upstream_list():
            name = path_to_name(path, onlybase=onlybase)
            yield git.UpstreamGitRepo(name, self.url_for(name, path), self.repo_type())

    def missing_repos(self):
        local_names = [repo.name for repo in self.local_repos()]
        for repo in self.upstream_repos():
            if repo.name not in local_names:
                yield repo

    def repo_type(self):
        repo_type = self.config["clone"].get("type", "mirror")
        if repo_type not in ("mirror", "mirror-ff", "bare", "working"):
            raise ValueError("Invalid type: %r" % repo_type)  # pragma: no cover
        return repo_type

    def warn_of_new_clone(self):
        return self.config["clone"].getboolean("warn", True)
